from dataclasses import dataclass


@dataclass
class Command:
    name: str
    url: str = 'remote'


COMMAND_ENGINE_ON = Command(name='REON')
COMMAND_ENGINE_OFF = Command(name='REOFF')
COMMAND_COMFORT_ON = Command(name='ROCOMFORTON')
COMMAND_COMFORT_OFF = Command(name='ROCOMFORTOFF')
COMMAND_HVAC_ON = Command(name='ROHVACON')
COMMAND_HVAC_OFF = Command(name='ROHVACOFF')
COMMAND_PRECOND_ON = Command(name='ROPRECOND')
COMMAND_PRECOND_OFF = Command(name='ROPRECOND_OFF')
COMMAND_LIGHTS_HORN = Command(name='HBLF')
COMMAND_LIGHTS = Command(name='ROLIGHTS')
COMMAND_DOORS_UNLOCK = Command(name='RDU')
COMMAND_DOORS_LOCK = Command(name='RDL')
COMMAND_TRUNK_UNLOCK = Command(name='ROTRUNKUNLOCK')
COMMAND_TRUNK_LOCK = Command(name='ROTRUNKLOCK')
COMMAND_LIFTGATE_UNLOCK = Command(name='ROLIFTGATEUNLOCK')
COMMAND_LIFTGATE_LOCK = Command(name='ROLIFTGATELOCK')
COMMAND_CHARGE = Command(name='CNOW', url='ev/chargenow')
COMMAND_DEEP_REFRESH = Command(name='DEEPREFRESH', url='ev')
COMMAND_REFRESH_LOCATION = Command(name='VF', url='location')

COMMANDS = [
    COMMAND_ENGINE_ON,
    COMMAND_ENGINE_OFF,
    COMMAND_COMFORT_ON,
    COMMAND_COMFORT_OFF,
    COMMAND_HVAC_ON,
    COMMAND_HVAC_OFF,
    COMMAND_PRECOND_ON,
    COMMAND_PRECOND_OFF,
    COMMAND_LIGHTS_HORN,
    COMMAND_LIGHTS,
    COMMAND_DOORS_UNLOCK,
    COMMAND_DOORS_LOCK,
    COMMAND_TRUNK_UNLOCK,
    COMMAND_TRUNK_LOCK,
    COMMAND_LIFTGATE_UNLOCK,
    COMMAND_LIFTGATE_LOCK,
    COMMAND_CHARGE,
    COMMAND_DEEP_REFRESH,
    COMMAND_REFRESH_LOCATION,
]

COMMANDS_BY_NAME = {x.name: x for x in COMMANDS}



import requests
import uuid
import json
import boto3
import base64
import logging
import http.client as http_client

from datetime import datetime, timedelta
from requests_auth_aws_sigv4 import AWSSigV4

from .command import Command
from .brands import Brand


class API:
    def __init__(self, email: str, password: str, pin: str, brand: Brand, disable_tls_verification: bool = False, dev_mode: bool = False, debug: bool = False):
        self.email = email
        self.password = password
        self.pin = pin
        self.brand = brand
        self.dev_mode = dev_mode

        self.uid: str = None
        self.aws_auth: AWSSigV4 = None

        self.sess = requests.Session()
        self.cognito_client = None

        self.expire_time: datetime = None

        if disable_tls_verification:
            self.sess.verify = False
            import urllib3
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        if debug:
            http_client.HTTPConnection.debuglevel = 1
            logging.basicConfig()
            logging.getLogger().setLevel(logging.DEBUG)
            requests_log = logging.getLogger('requests.packages.urllib3')
            requests_log.setLevel(logging.DEBUG)
            requests_log.propagate = True

    def _with_default_params(self, params: dict):
        return params | {
            'targetEnv': 'jssdk',
            'loginMode': 'standard',
            'sdk': 'js_latest',
            'authMode': 'cookie',
            'sdkBuild': '12234',
            'format': 'json',
            'APIKey': self.brand.login_api_key,
        }

    def _default_aws_headers(self, key: str):
        return {
            'x-clientapp-name': 'CWP',
            'x-clientapp-version': '1.0',
            'clientrequestid': uuid.uuid4().hex.upper()[0:16],
            'x-api-key': key,
            'locale': self.brand.locale,
            'x-originator-type': 'web',
        }

    def set_tls_verification(self, verify: bool):
        self.sess.verify = verify

    def login(self):
        '''Logs into the Uconnect and caches the auth tokens'''

        if self.cognito_client is None:
            self.cognito_client = boto3.client(
                'cognito-identity', self.brand.region)

        r = self.sess.request(
            method='GET',
            url=self.brand.login_url + '/accounts.webSdkBootstrap',
            params={'apiKey': self.brand.login_api_key}
        )

        r.raise_for_status()
        r = r.json()

        if r['statusCode'] != 200:
            raise Exception(f'bootstrap failed: {r}')

        r = self.sess.request(
            method='POST',
            url=self.brand.login_url + '/accounts.login',
            params=self._with_default_params({
                'loginID': self.email,
                'password': self.password,
                'sessionExpiration': 300,
                'include': 'profile,data,emails,subscriptions,preferences'
            })
        )

        r.raise_for_status()
        r = r.json()

        if r['statusCode'] != 200:
            raise Exception(f'account login failed: {r}')

        self.uid = r['UID']
        login_token = r['sessionInfo']['login_token']

        r = self.sess.request(
            method='POST',
            url=self.brand.login_url + '/accounts.getJWT',
            params=self._with_default_params({
                'login_token': login_token,
                'fields': 'profile.firstName,profile.lastName,profile.email,country,locale,data.disclaimerCodeGSDP'
            })
        )

        r.raise_for_status()
        r = r.json()

        if r['statusCode'] != 200:
            raise Exception(f'unable to obtain JWT: {r}')

        exc = None
        for url in self.brand.token_url:
            try:
                r = self.sess.request(
                    method='POST',
                    url=url,
                    headers=self._default_aws_headers(
                        self.brand.api.key) | {'content-type': 'application/json'},
                    json={'gigya_token': r['id_token']}
                )
            except Exception as e:
                exc = e
            else:
                break
        else:
            raise Exception(f'unable to obtain token: {exc}')

        r.raise_for_status()
        r = r.json()

        token = r.get('Token', None)
        identity_id = r.get('IdentityId', None)
        if token is None or identity_id is None:
            raise Exception(f'unable to obtain identity & token: {r}')

        r = self.cognito_client.get_credentials_for_identity(
            IdentityId=identity_id,
            Logins={'cognito-identity.amazonaws.com': token},
        )

        creds = r.get('Credentials', None)
        if not creds:
            raise Exception(f'unable to obtain AWS credentials: {r}')

        self.aws_auth = AWSSigV4(
            'execute-api',
            region=self.brand.region,
            aws_access_key_id=creds['AccessKeyId'],
            aws_secret_access_key=creds['SecretKey'],
            aws_session_token=creds['SessionToken'],
        )

        self.expire_time = creds['Expiration']

    def _refresh_token_if_needed(self):
        '''Checks if token is available and fresh, refreshes it otherwise'''

        if self.dev_mode:
            return

        if self.expire_time is None or datetime.now().astimezone() > self.expire_time - timedelta(minutes=5):
            try:
                self.login()
            except Exception as e:
                raise Exception(f'unable to login: {e}')

    def list_vehicles(self) -> list[dict]:
        '''Loads a list of vehicles with general info'''

        if self.dev_mode:
            with open('test_list.json') as f:
                return json.load(f)['vehicles']

        self._refresh_token_if_needed()

        r = self.sess.request(
            method='GET',
            url=self.brand.api.url + f'/v4/accounts/{self.uid}/vehicles',
            headers=self._default_aws_headers(
                self.brand.api.key) | {'content-type': 'application/json'},
            params={'stage': 'ALL'},
            auth=self.aws_auth,
        ).json()

        if not 'vehicles' in r:
            raise Exception(f"incorrect response: {r}")

        return r['vehicles']

    def get_vehicle(self, vin: str) -> dict:
        '''Gets a more detailed info abount a vehicle with a given VIN'''

        if self.dev_mode:
            with open(f'test_vehicle_{vin}.json') as f:
                return json.load(f)

        self._refresh_token_if_needed()

        return self.sess.request(
            method='GET',
            url=self.brand.api.url + f'/v2/accounts/{self.uid}/vehicles/{vin}/status',
            headers=self._default_aws_headers(
                self.brand.api.key) | {'content-type': 'application/json'},
            auth=self.aws_auth,
        ).json()

    def get_vehicle_status(self, vin: str) -> dict:
        '''Loads another part of status of a vehicle with a given VIN'''

        if self.dev_mode:
            with open(f'test_vehicle_status_{vin}.json') as f:
                return json.load(f)

        self._refresh_token_if_needed()

        return self.sess.request(
            method='GET',
            url=self.brand.api.url +
            f'/v1/accounts/{self.uid}/vehicles/{vin}/remote/status',
            headers=self._default_aws_headers(
                self.brand.api.key) | {'content-type': 'application/json'},
            auth=self.aws_auth,
        ).json()

    def get_vehicle_location(self, vin: str) -> dict:
        '''Gets last known location of a vehicle with a given VIN'''

        if self.dev_mode:
            with open(f'test_vehicle_location_{vin}.json') as f:
                return json.load(f)

        self._refresh_token_if_needed()

        return self.sess.request(
            method='GET',
            url=self.brand.api.url +
            f'/v1/accounts/{self.uid}/vehicles/{vin}/location/lastknown',
            headers=self._default_aws_headers(
                self.brand.api.key) | {'content-type': 'application/json'},
            auth=self.aws_auth,
        ).json()

    def get_vehicle_notifications(self, vin: str, limit: int | None = 30) -> dict:
        '''Loads notifications for a vehicle with a given VIN'''

        self._refresh_token_if_needed()

        url = self.brand.api.url + \
            f'/v1/accounts/{self.uid}/vehicles/{vin}/notifications'

        if limit is not None:
            url += f'?limit={limit}'

        return self.sess.request(
            method='GET',
            url=url,
            headers=self._default_aws_headers(
                self.brand.api.key) | {'content-type': 'application/json'},
            auth=self.aws_auth,
        ).json()

    def command(self,
                vin: str, cmd: Command):
        '''Sends given command to the vehicle with a given VIN'''

        if self.dev_mode:
            return

        data = {
            'pin': base64.b64encode(self.pin.encode()).decode(encoding='utf-8'),
        }

        self._refresh_token_if_needed()

        exc = None
        for auth in self.brand.auth:
            try:
                r = self.sess.request(
                    method='POST',
                    url=auth.url +
                    f'/v1/accounts/{self.uid}/ignite/pin/authenticate',
                    headers=self._default_aws_headers(auth.token) | {
                        'content-type': 'application/json'},
                    auth=self.aws_auth,
                    json=data,
                ).json()
            except Exception as e:
                exc = e
            else:
                break
        else:
            raise Exception(f'Authentication failed: {exc}')

        if not 'token' in r:
            raise Exception(f'authentication failed: no token found: {r}')

        data = {
            'command': cmd.name,
            'pinAuth': r['token'],
        }

        r = self.sess.request(
            method='POST',
            url=self.brand.api.url +
            f'/v1/accounts/{self.uid}/vehicles/{vin}/{cmd.url}',
            headers=self._default_aws_headers(
                self.brand.api.key) | {'content-type': 'application/json'},
            auth=self.aws_auth,
            json=data,
        ).json()

        if not 'responseStatus' in r or r['responseStatus'] != 'pending':
            error = r.get('debugMsg', 'unknown error')
            raise Exception(f'command queuing failed: {error}')

        return r['correlationId']



from dataclasses import dataclass

API_URL_EU = 'https://channels.sdpr-01.fcagcv.com'
API_URL_US = 'https://channels.sdpr-02.fcagcv.com'

TOKEN_URL_EU = 'https://authz.sdpr-01.fcagcv.com/v2/cognito/identity/token'
TOKEN_URL_US = 'https://authz.sdpr-02.fcagcv.com/v2/cognito/identity/token'
TOKEN_URL_US_PREP = 'https://authz.sdpr-02.prep.fcagcv.com/v2/cognito/identity/token'

LOCALE_EU = 'de_de'
LOCALE_US = 'en_us'

REGION_US = 'us-east-1'
REGION_EU = 'eu-west-1'


@dataclass
class API:
    url: str
    key: str


API_EU = API(
    url='https://channels.sdpr-01.fcagcv.com',
    key='2wGyL6PHec9o1UeLPYpoYa1SkEWqeBur9bLsi24i',
)

API_US = API(
    url='https://channels.sdpr-02.fcagcv.com',
    key='OgNqp2eAv84oZvMrXPIzP8mR8a6d9bVm1aaH9LqU',
)


@dataclass
class Auth:
    url: str
    token: str


AUTH_EU = Auth(
    url='https://mfa.fcl-01.fcagcv.com',
    token='JWRYW7IYhW9v0RqDghQSx4UcRYRILNmc8zAuh5ys',
)

AUTH_US = Auth(
    url='https://mfa.fcl-02.fcagcv.com',
    token='fNQO6NjR1N6W0E5A6sTzR3YY4JGbuPv48Nj9aZci',
)

AUTH_US_PREP = Auth(
    url='https://mfa.fcl-02.prep.fcagcv.com',
    token='lHBEtsqT1Y5oKvzhvA9KW6rkirU3ZtGf44jTIiQV'
)


@dataclass
class Brand:
    name: str
    region: str
    login_api_key: str
    login_url: str
    token_url: list[str]
    api: API
    auth: list[Auth]
    locale: str


FIAT_EU = Brand(
    name='FIAT_EU',
    region=REGION_EU,
    login_api_key='3_mOx_J2dRgjXYCdyhchv3b5lhi54eBcdCTX4BI8MORqmZCoQWhA0mV2PTlptLGUQI',
    login_url='https://loginmyuconnect.fiat.com',
    token_url=[TOKEN_URL_EU, TOKEN_URL_US],
    api=API_EU,
    auth=[AUTH_EU, AUTH_US],
    locale=LOCALE_EU,
)

FIAT_US = Brand(
    name='FIAT_US',
    region=REGION_US,
    login_api_key='3_WfFvlZJwcSdOD0LFQCngUV3W390R4Yshpuq3RsZvnV4VG0c9Q6R0RtDwcXc8dTrI',
    login_url='https://login-us.fiat.com',
    token_url=[TOKEN_URL_US, TOKEN_URL_EU],
    api=API_US,
    auth=[AUTH_EU, AUTH_US],
    locale=LOCALE_US,
)

FIAT_ASIA = Brand(
    name='FIAT_ASIA',
    region=REGION_EU,
    login_api_key='4_YAQNaPqdPEUbbzhvhunKAA',
    login_url='https://login-iap.fiat.com',
    token_url=[TOKEN_URL_EU, TOKEN_URL_US],
    api=API_EU,
    auth=[AUTH_EU, AUTH_US],
    locale=LOCALE_EU,
)

FIAT_CANADA = Brand(
    name='FIAT_CANADA',
    region=REGION_US,
    login_api_key='3_Ii2kSgQm4ljy19LIZeLwa76OlmWbpSa8w3aSP5VJdx19tub3oWxsFR-HEusDnUEh',
    login_url='https://login-stage-us.fiat.com',
    token_url={TOKEN_URL_US_PREP, TOKEN_URL_US, TOKEN_URL_EU},
    api=API_US,
    auth=[AUTH_US_PREP, AUTH_US, AUTH_EU],
    locale=LOCALE_US,
)

ALFA_ROMEO_US_CANADA = Brand(
    name='ALFA_ROMEO_US_CANADA',
    region=REGION_US,
    login_api_key='3_FSxGyaktviayTDRcgp9r9o2KjuFSrHT13wWNN9zPrvAGUCoXPDqoIPOwlBUhck4A',
    login_url='https://login-us.alfaromeo.com',
    token_url=[TOKEN_URL_US, TOKEN_URL_EU],
    api=API_US,
    auth=[AUTH_US, AUTH_EU],
    locale=LOCALE_US,
)

ALFA_ROMEO_ASIA = Brand(
    name='ALFA_ROMEO_ASIA',
    region=REGION_EU,
    login_api_key='4_PSQeADnQ4p5XOaDgT0B5pA',
    login_url='https://login-iap.alfaromeo.com',
    token_url=[TOKEN_URL_EU, TOKEN_URL_US],
    api=API_EU,
    auth=[AUTH_EU, AUTH_US],
    locale=LOCALE_EU,
)

ALFA_ROMEO_EU = Brand(
    name='ALFA_ROMEO_EU',
    region=REGION_EU,
    login_api_key='3_h8sj2VQI-KYXiunPq9a1QuAA4yWkY0r5AD1u8A8B1RPn_Cvl54xcoc2-InH5onJ1',
    login_url='https://login.alfaromeo.com',
    token_url=[TOKEN_URL_EU, TOKEN_URL_US],
    api=API_EU,
    auth=[AUTH_EU, AUTH_US],
    locale=LOCALE_EU,
)

CHRYSLER_CANADA = Brand(
    name='CHRYSLER_CANADA',
    region=REGION_US,
    login_api_key='3_gdhu-ur4jc2hEryDMnF4YPELkjzSi-invZTjop4isZu4ReHodVcuL44u93cOUqMC',
    login_url='https://login-stage-us.chrysler.com',
    token_url=[TOKEN_URL_US, TOKEN_URL_EU],
    api=API_US,
    auth=[AUTH_US, AUTH_EU],
    locale=LOCALE_US,
)

CHRYSLER_US = Brand(
    name='CHRYSLER_US',
    region=REGION_US,
    login_api_key='3_cv4AzHkJh48-cqwaf_Ahcg1HnsmQqz1lm0sOdVdHW5FjT3m6SyywywOBaskBQqwn',
    login_url='https://login-us.chrysler.com',
    token_url=[TOKEN_URL_US, TOKEN_URL_EU],
    api=API_US,
    auth=[AUTH_US, AUTH_EU],
    locale=LOCALE_US,
)

MASERATI_EU = Brand(
    name='MASERATI_EU',
    region=REGION_EU,
    login_api_key='3_rNbVuhn2gIt3BnLjlGsJcMo26Lft3avDne_FLRT34Dy_9OxHtCVOnplwY436lGZa',
    login_url='https://login.maserati.com',
    token_url=[TOKEN_URL_EU, TOKEN_URL_US],
    api=API_EU,
    auth=[AUTH_EU, AUTH_US],
    locale=LOCALE_EU,
)

MASERATI_ASIA = Brand(
    name='MASERATI_ASIA',
    region=REGION_EU,
    login_api_key='4_uwF-in6KF-aMbEkPAb-fOg',
    login_url='https://accounts.au1.gigya.com',
    token_url=[TOKEN_URL_EU, TOKEN_URL_US],
    api=API_EU,
    auth=[AUTH_EU, AUTH_US],
    locale=LOCALE_EU,
)

MASERATI_US_CANADA = Brand(
    name='MASERATI_US_CANADA',
    region=REGION_US,
    login_api_key='3_nShL4-O7IL0OGqroO8AzwiRU0-ZHcBZ4TLBrh5MORusMo5XYxhCLXPYfjI4OOLOy',
    login_url='https://login-us.maserati.com',
    token_url=[TOKEN_URL_US, TOKEN_URL_EU],
    api=API_US,
    auth=[AUTH_US, AUTH_EU],
    locale=LOCALE_US,
)

JEEP_EU = Brand(
    name='JEEP_EU',
    region=REGION_EU,
    login_api_key='3_ZvJpoiZQ4jT5ACwouBG5D1seGEntHGhlL0JYlZNtj95yERzqpH4fFyIewVMmmK7j',
    login_url='https://login.jeep.com',
    token_url=[TOKEN_URL_EU, TOKEN_URL_US],
    api=API_EU,
    auth=[AUTH_EU, AUTH_US],
    locale=LOCALE_EU,
)

JEEP_US = Brand(
    name='JEEP_US',
    region=REGION_US,
    login_api_key='3_5qxvrevRPG7--nEXe6huWdVvF5kV7bmmJcyLdaTJ8A45XUYpaR398QNeHkd7EB1X',
    login_url='https://login-us.jeep.com',
    token_url=[TOKEN_URL_US, TOKEN_URL_EU],
    api=API_US,
    auth=[AUTH_US, AUTH_EU],
    locale=LOCALE_US,
)

JEEP_ASIA = Brand(
    name='JEEP_ASIA',
    region=REGION_EU,
    login_api_key='4_zqGYHC7rM8RCHHl4YFDebA',
    login_url='https://login-iap.jeep.com',
    token_url=[TOKEN_URL_EU, TOKEN_URL_US],
    api=API_EU,
    auth=[AUTH_EU, AUTH_US],
    locale=LOCALE_EU,
)

DODGE_US = Brand(
    name='DODGE_US',
    region=REGION_US,
    login_api_key='4_dSRvo6ZIpp8_St7BF9VHGA',
    login_url='https://login-us.dodge.com',
    token_url=[TOKEN_URL_US, TOKEN_URL_EU],
    api=API_US,
    auth=[AUTH_US, AUTH_EU],
    locale=LOCALE_US,
)

RAM_US = Brand(
    name='RAM_US',
    region=REGION_US,
    login_api_key='3_7YjzjoSb7dYtCP5-D6FhPsCciggJFvM14hNPvXN9OsIiV1ujDqa4fNltDJYnHawO',
    login_url='https://login-us.ramtrucks.com',
    token_url=[TOKEN_URL_US, TOKEN_URL_EU],
    api=API_US,
    auth=[AUTH_US, AUTH_EU],
    locale=LOCALE_US,
)

BRANDS = {
    FIAT_EU.name: FIAT_EU,
    FIAT_US.name: FIAT_US,
    FIAT_CANADA.name: FIAT_CANADA,
    FIAT_ASIA.name: FIAT_ASIA,

    JEEP_EU.name: JEEP_EU,
    JEEP_US.name: JEEP_US,
    JEEP_ASIA.name: JEEP_ASIA,

    DODGE_US.name: DODGE_US,

    RAM_US.name: RAM_US,

    CHRYSLER_CANADA.name: CHRYSLER_CANADA,
    CHRYSLER_US.name: CHRYSLER_US,

    ALFA_ROMEO_US_CANADA.name: ALFA_ROMEO_US_CANADA,
    ALFA_ROMEO_EU.name: ALFA_ROMEO_EU,
    ALFA_ROMEO_ASIA.name: ALFA_ROMEO_ASIA,

    MASERATI_ASIA.name: MASERATI_ASIA,
    MASERATI_EU.name: MASERATI_EU,
    MASERATI_US_CANADA.name: MASERATI_US_CANADA,
}



from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
from typing import Dict
from datetime import datetime, timedelta
from time import sleep

from .api import API
from .brands import Brand
from .command import Command, COMMANDS_BY_NAME


def convert(v):
    if not isinstance(v, str):
        return v

    if v == 'null':
        return None

    try:
        v = int(v)
    except:
        try:
            v = float(v)
        except:
            pass

    return v


def sg(dct: dict, *keys):
    if not isinstance(dct, dict):
        return None

    for key in keys:
        try:
            dct = dct[key]
        except KeyError:
            return None

    return convert(dct)


def sg_eq(dct: dict, expect, *keys):
    v = sg(dct, *keys)

    if v is None:
        return None

    return v == expect


CHARGING_LEVELS = {
    'DEFAULT': 0,
    'LEVEL_1': 1,
    'LEVEL_2': 2,
    'LEVEL_3': 3,
}

CHARGING_LEVEL_PREFS = {
    'LEVEL_ONE': 1,
    'LEVEL_TWO': 2,
    'LEVEL_THREE': 3,
    'LEVEL_FOUR': 4,
    'LEVEL_FIVE': 5,
}


@dataclass_json
@dataclass
class Location:
    longitude: float = None
    latitude: float = None
    altitude: float = None
    bearing: float = None
    is_approximate: bool = None
    updated: datetime = None

    def __repr__(self):
        return f'lat: {self.latitude}, lon: {self.longitude} (updated {self.updated})'


@dataclass_json
@dataclass
class Vehicle:
    vin: str

    # General info
    nickname: str
    make: str
    model: str
    year: str
    region: str

    # Status
    ignition_on: bool = None
    trunk_locked: bool = None

    odometer: float = None
    odometer_unit: str = None
    days_to_service: int = None
    distance_to_service: int = None
    distance_to_service_unit: str = None
    distance_to_empty: int = None
    distance_to_empty_unit: str = None
    battery_voltage: float = None
    oil_level: int = None
    fuel_low: bool = None
    fuel_amount: int = None

    # EV related
    plugged_in: bool = None
    ev_running: bool = None
    charging: bool = None
    charging_level: int = None
    charging_level_preference: int = None
    state_of_charge: int = None
    time_to_fully_charge_l3: int = None
    time_to_fully_charge_l2: int = None

    # Wheels
    wheel_front_left_pressure: float = None
    wheel_front_left_pressure_unit: str = None
    wheel_front_left_pressure_warning: bool = None
    wheel_front_right_pressure: float = None
    wheel_front_right_pressure_unit: str = None
    wheel_front_right_pressure_warning: bool = None
    wheel_rear_left_pressure: float = None
    wheel_rear_left_pressure_unit: str = None
    wheel_rear_left_pressure_warning: bool = None
    wheel_rear_right_pressure: float = None
    wheel_rear_right_pressure_unit: str = None
    wheel_rear_right_pressure_warning: bool = None

    # Doors
    door_driver_locked: bool = None
    door_passenger_locked: bool = None
    door_rear_left_locked: bool = None
    door_rear_right_locked: bool = None

    # Windows
    window_driver_closed: bool = None
    window_passenger_closed: bool = None

    location: Location = None
    supported_commands: list[str] = field(default_factory=list)
    last_full_update: datetime = None

    def __repr__(self):
        return f'{self.vin} (nick: {self.nickname})'


def _update_vehicle(v: Vehicle, p: dict) -> Vehicle:
    vi = sg(p, 'vehicleInfo')
    ev = sg(p, 'evInfo')
    batt = sg(ev, 'battery')

    v.battery_voltage = sg(vi, 'batteryInfo', 'batteryVoltage', 'value')
    v.charging = sg_eq(batt, 'CHARGING', 'chargingStatus')
    v.charging_level = CHARGING_LEVELS.get(sg(batt, 'chargingLevel'), None)
    v.charging_level_preference = CHARGING_LEVEL_PREFS.get(
        sg(ev, 'chargePowerPreference'), None)
    v.plugged_in = sg(batt, 'plugInStatus')
    v.state_of_charge = sg(batt, 'stateOfCharge')

    v.days_to_service = sg(vi, 'daysToService')
    v.distance_to_service = sg(
        vi, 'distanceToService', 'distanceToService', 'value')
    v.distance_to_service_unit = sg(
        vi, 'distanceToService', 'distanceToService', 'unit')
    v.distance_to_empty = sg(vi, 'fuel', 'distanceToEmpty', 'value')
    v.distance_to_empty_unit = sg(vi, 'fuel', 'distanceToEmpty', 'unit')
    v.fuel_low = sg(vi, 'fuel', 'isFuelLevelLow')
    v.fuel_amount = sg(vi, 'fuel', 'fuelAmountLevel')
    v.oil_level = sg(vi, 'oilLevel', 'oilLevel')

    v.ignition_on = sg_eq(ev, 'ON', 'ignitionStatus')
    v.time_to_fully_charge_l3 = sg(batt, 'timeToFullyChargeL3')
    v.time_to_fully_charge_l2 = sg(batt, 'timeToFullyChargeL2')
    v.odometer = sg(vi, 'odometer', 'odometer', 'value')
    v.odometer_unit = sg(vi, 'odometer', 'odometer', 'unit')

    if isinstance(vi, dict) and 'tyrePressure' in vi:
        tp = {x['type']: x for x in vi['tyrePressure']}

        v.wheel_front_left_pressure = sg(tp, 'FL', 'pressure', 'value')
        v.wheel_front_left_pressure_unit = sg(tp, 'FL', 'pressure', 'unit')
        v.wheel_front_left_pressure_warning = sg(tp, 'FL', 'warning')

        v.wheel_front_right_pressure = sg(tp, 'FR', 'pressure', 'value')
        v.wheel_front_right_pressure_unit = sg(tp, 'FR', 'pressure', 'unit')
        v.wheel_front_right_pressure_warning = sg(tp, 'FR', 'warning')

        v.wheel_rear_left_pressure = sg(tp, 'RL', 'pressure', 'value')
        v.wheel_rear_left_pressure_unit = sg(tp, 'RL', 'pressure', 'unit')
        v.wheel_rear_left_pressure_warning = sg(tp, 'RL', 'warning')

        v.wheel_rear_right_pressure = sg(tp, 'RR', 'pressure', 'value')
        v.wheel_rear_right_pressure_unit = sg(tp, 'RR', 'pressure', 'unit')
        v.wheel_rear_right_pressure_warning = sg(tp, 'RR', 'warning')

    return v


class Client:
    def __init__(self, email: str, password: str, pin: str, brand: Brand, disable_tls_verification: bool = False, dev_mode: bool = False, debug: bool = False):
        self.api = API(email, password, pin, brand,
                       disable_tls_verification=disable_tls_verification, dev_mode=dev_mode, debug=debug)
        self.vehicles: Dict[str, Vehicle] = {}

    def set_tls_verification(self, verify: bool):
        '''Enable or disable TLS certificate verification'''

        self.api.set_tls_verification(verify)

    def refresh(self):
        '''Refreshes all the vehicle data and caches it locally'''

        vehicles = self.api.list_vehicles()

        for x in vehicles:
            full_update_done = True
            vin = x['vin']

            if not vin in self.vehicles:
                vehicle = Vehicle(vin=vin, nickname=sg(x, 'nickname'), make=sg(x, 'make'), model=sg(
                    x, 'modelDescription'), year=sg(x, 'tsoModelYear'), region=sg(x, 'soldRegion'))
                self.vehicles[vin] = vehicle
            else:
                vehicle = self.vehicles[vin]

            info = self.api.get_vehicle(vin)
            if not set(['vehicleInfo', 'evInfo']).intersection(info.keys()):
                full_update_done = False

            _update_vehicle(vehicle, info)

            try:
                loc = self.api.get_vehicle_location(vin)

                vehicle.location = Location(
                    longitude=sg(loc, 'longitude'),
                    latitude=sg(loc, 'latitude'),
                    altitude=sg(loc, 'altitude'),
                    bearing=sg(loc, 'bearing'),
                    is_approximate=sg(loc, 'isLocationApprox'),
                    updated=datetime.fromtimestamp(loc['timeStamp'] / 1000)
                )
            except:
                full_update_done = False

            try:
                s = self.api.get_vehicle_status(vin)

                # Check at least one key is present in the response
                if not set(['doors', 'windows', 'trunk', 'evRunning']).intersection(set(s.keys())):
                    full_update_done = False

                if 'doors' in s:
                    vehicle.door_driver_locked = sg_eq(
                        s, 'LOCKED', 'doors', 'driver', 'status')
                    vehicle.door_passenger_locked = sg_eq(
                        s, 'LOCKED', 'doors', 'passenger', 'status')
                    vehicle.door_rear_left_locked = sg_eq(
                        s, 'LOCKED', 'doors', 'leftRear', 'status')
                    vehicle.door_rear_right_locked = sg_eq(
                        s, 'LOCKED', 'doors', 'rightRear', 'status')

                if 'windows' in s:
                    vehicle.window_driver_closed = sg_eq(
                        s, 'CLOSED', 'windows', 'driver', 'status')
                    vehicle.window_passenger_closed = sg_eq(
                        s, 'CLOSED', 'windows', 'passenger', 'status')

                vehicle.trunk_locked = sg_eq(s, 'LOCKED', 'trunk', 'status')
                vehicle.ev_running = sg_eq(s, 'ON', 'evRunning', 'status')
            except:
                full_update_done = False

            enabled_services = []
            if 'services' in x:
                enabled_services = [v['service'] for v in x['services']
                                    if sg(v, 'vehicleCapable') and sg(v, 'serviceEnabled')]

            vehicle.supported_commands = [
                v for v in enabled_services if v in COMMANDS_BY_NAME]

            if full_update_done:
                vehicle.last_full_update = datetime.now().astimezone()

    def get_vehicles(self):
        '''Returns all vehicles data. Must execute refresh method before.'''

        return self.vehicles

    def command(self, vin: str, cmd: Command):
        '''Execute a given command against a car with a given VIN'''

        return self.api.command(vin, cmd)

    def _get_commands_statuses(self, vin: str) -> dict:
        r = self.api.get_vehicle_notifications(vin)

        return {
            x['correlationId']: (x['notification']['data']
                                 ['status'].lower() == "success")
            for x in r['notifications']['items']
        }

    def command_verify(self, vin: str, cmd: Command,
                       timeout: timedelta = timedelta(seconds=60),
                       interval: timedelta = timedelta(seconds=2)):
        '''Execute a given command against a car with a given VIN and poll for the status'''

        id = self.command(vin, cmd)

        start = datetime.now()
        while datetime.now() - start < timeout:
            sleep(interval.seconds)
            r = self._get_commands_statuses(vin)
            if id in r:
                return r[id]

        raise Exception(f'unable to obtain command status: timed out (id {id})')
